#!/usr/bin/python
#  calcvalues.py                     RBLandau 20080913
# 
# Copyright (c) 2008, 2009, Dell, Inc. 
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
# 
#     * Redistributions of source code must retain the above
# copyright notice, this list of conditions and the following
# disclaimer.
# 
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided
# with the distribution.
# 
#     * Neither the name of the Dell, Inc., nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# 
# These are the functions that are used to calculate the values of 
# CIM properties.  They will be called by name from the getValue
# methods of the CimProp instances.  They depend on the bin full
# of SNMP variables already having been retrieved.  
# 
# This is the heart of the process to evaluate CIM properties, 
# as you can tell from its size and complexity.  There are several 
# types of functions here:
#  - Actual calculation functions that produce an output value for
#    a specific property;
#  - Calculation functions that are generally designed to produce 
#    values for a variety of similar properties across classes.
#  - Pieces-parts functions that do part of the calculation 
#    process.  
# 
# Functions are clled with two arguments:
#  - A ref to the property instance for which we are trying to 
#    calculate the value; and 
#  - Optionally a seonc arg, default value null string, that contains
#    the name of an snmp variable that is already stored in the 
#    property instance's dictionary, or a constant value string for
#    some other purpose, usually to be returned directly as the 
#    property value.  

from NewTrace import TRC

from dicts      import *
from datastore  import *
from sys        import modules
import time
import re

# Oops, this one has to be done on demand only when needed.
# Namespace caching causes a problem otherwise.  
#from cimprop    import CCimProp

# There is a *lot* of evolutionary code here.  You will find that the 
# same function may be done in two or three different ways.  
# Explanation: I did it one way early on, usually brute force.  Then 
# later on I discovered an easier, more elegant, or simply shorter 
# way to accomplish the same task.  So later versions of the same 
# function may be coded in a very different way.  
# Excuse: I didn't go back to fix the older versions just because
# I'm lazy and didn't have the time.  It's a   p r o t o t y p e .
# It's okay if the East German judge wants to dock me some points 
# for imperfectly elegant coding.  


#=========== G e n e r a l   m a p p i n g s

# f D i r e c t ( )
#
# The special mapping function that takes an SNMP variable value
# and puts it directly into a CIM property.  
# 
# Strangeness here; also done in fCapacity.
# If the property is an integer, coerce the value to be non-negative.
# This is done to handle those odd variables in Printer MIB that 
# are really counters but use some out-of-band negative values to
# signify unknown, not implemented, unbounded, etc.  
#
# Beware the evil eval() that is used here and anywhere else that
# a value must be retrieved by name.  No other reasonable way to 
# substitute the string value of the name on the fly.  
# At least getStrVal() and getIntVal(), and probably others, too, 
# have the same problem.  
def fDirect(propinst,myvarname):
    temp = eval("propinst.mget(\"%s\")"%myvarname)
    try:
        propinst.value = int(temp)      # If it is an integer value, 
        if propinst.value < 0:          # make it a nonnegative value;
            propinst.value = 0
    except (ValueError,TypeError):
        if temp is None:
            propinst.value = 0
        else:
            propinst.value = temp       # else return the value as given.
    propinst.mset("value",propinst.value)
    TRC.tracef(2,"CALC","exit  calc fDirect inst|%s| name|%s| val|%s|" \
        % (propinst.sFullname,myvarname,propinst.value))
    return propinst.value

# f D i r e c t O r N u l l 
# 
# Special version for some declared measurements, such as 
# CIM_PrintMediaPath.MaxSpeed.  If the basis of the property value
# is not actual, then the value of the property shall be null.  
def fDirectOrNull(propinst,myvarname):
    temp = eval("propinst.mget(\"%s\")"%myvarname)
    try:
        propinst.value = int(temp)      # If it is an integer value, 
        if propinst.value < 0:          # if negative then return
            propinst.value = None       # null.  
    except (ValueError,TypeError):
        if temp is None:
            propinst.value = None
        else:
            propinst.value = temp       # else return the value as given.
    propinst.mset("value",propinst.value)
    TRC.tracef(2,"CALC","exit  calc fDirectOrNull inst|%s| name|%s| val|%s|" \
        % (propinst.sFullname,myvarname,propinst.value))
    return propinst.value


# f C o n s t a n t ( )
#
# The special mapping function that takes a constant value
# and puts it directly into a CIM property.  
def fConstant(propinst,myvalue):
    #propinst.value = dConstantprop2val.mget(propinst.sFullname)
    propinst.value = myvalue
    propinst.mset("value",propinst.value)
    TRC.tracef(2,"CALC","exit  calc fConstant inst|%s| invalue|%s| dictval|%s|" \
        % (propinst.sFullname,myvalue,propinst.value))
    return propinst.value

# f N u l l ( )
#
# The special mapping function that stores a null value.  
def fNull(propinst,myvar=""):
    propinst.value = None
    propinst.mset("value",propinst.value)
    propinst.mset("valuelist",None)
    TRC.tracef(2,"CALC","exit  calc fNull class|%s| val|%s|"\
        % (propinst.sFullname,propinst.value))
    return propinst.value

# f E n u m ( )
#    
# The special mapping function that maps from the SNMP enum list
# to corresponding CIM enum values.  
def fEnum(propinst,myvar=""):
    # TBS
    return propinst.value

# f D e p r e c a t e d ( )
#
# The special mapping function that stores a recognizable value for
# deprecated properties.  
# For the real version, change the stored value to None or a null string.
def fDeprecated(propinst,myvar=""):
    propinst.value = "((deprecated))"
    propinst.mset("value",propinst.value)
    propinst.mset("valuelist",propinst.value)
    TRC.tracef(2,"CALC","exit  calc fDeprecated class|%s| val|%s|"\
        % (propinst.sFullname,propinst.value))
    return propinst.value


#=========== S t a t u s   m a p p i n g s

# f A v a i l a b i l i t y S t a t u s ( )
#
# Extract the AvailabilityStatus bits from a SubUnitStatus variable.  
def fAvailabilityStatus(propinst,myvarname):
    #myvarname = dStatusprop2var.mget(propinst.sFullname)
    TRC.tracef(5,"CALC","enter calc fAvailabilityStatus inst|%s| name|%s| val|%s|" \
        % (propinst.sFullname,myvarname,propinst.value))

    temp = fGetIntVal(propinst,myvarname,5)
    # Could do this with a dictionary instead.  
    temp = temp & 7
    if temp ==   0:             # available idle
      result = 3
    elif temp == 1:             # unavailable on request
      result = 7
    elif temp == 2:             # available standby
      result = 4
    elif temp == 3:             # unavailable broken
      result = 8
    elif temp == 4:             # available active
      result = 5
    elif temp == 5:             # unknown
      result = 2
    elif temp == 6:             # available busy
      result = 6
    else:                       # other
      result = 1
    return fCommonExit(propinst,myvarname,result,"fAvailabilityStatus")

#    else:
#        result = 2                  # unknown if not there at all
    propinst.value = result
    propinst.mset("value",propinst.value)
    TRC.tracef(2,"CALC","exit  calc fAvailabilityStatus inst|%s| name|%s| val|%s|" \
        % (propinst.sFullname,myvarname,propinst.value))
    return propinst.value

# f C r i t i c a l S t a t u s ( )
#
def fCriticalStatus(propinst,myvarname):
    temp = fGetIntVal(propinst,myvarname,0)
    result = temp & 16              # Get only the bit for CriticalStatus
    sresult = fMakeBoolean(result)
    return fCommonExit(propinst,myvarname,sresult,"fCriticalStatus")

# f N o n c r i t i c a l S t a t u s ( )
#
def fNoncriticalStatus(propinst,myvarname):
    temp = fGetIntVal(propinst,myvarname,0)
    result = temp & 8               # Get only the bit for noncritical status
    sresult = fMakeBoolean(result)
    return fCommonExit(propinst,myvarname,sresult,"fNoncriticalStatus")
    
# f P r i m a r y S t a t u s ( )
#
# Extract the PrimaryStatus from a SubUnitStatus variable.  
def fPrimaryStatus(propinst,myvarname):
    #myvarname = dStatusprop2var.mget(propinst.sFullname)
    TRC.tracef(5,"CALC","enter calc fPrimaryStatus inst|%s| name|%s| val|%s|" \
        % (propinst.sFullname,myvarname,propinst.value))

    temp = fGetIntVal(propinst,myvarname,5)
    tempavail   = temp & 7
    tempnoncrit = temp & 8
    tempcrit    = temp & 16
    tempoffline = temp & 32
    if (tempavail&1==0):
      result = 1                # Ok
    elif (tempnoncrit) or (tempoffline):
      result = 3                # Degraded
    elif (tempcrit) or (tempavail==1) or (tempavail==3):
      result = 3                # Error
    elif (tempavail==5):
      result = 0                # Unknown
    return fCommonExit(propinst,myvarname,result,"fPrimaryStatus")

# f D e t a i l e d S t a t u s ( )
#
# Extract the DetailedStatus from a SubUnitStatus variable.  
def fDetailedStatus(propinst,myvarname):
    #myvarname = dStatusprop2var.mget(propinst.sFullname)
    TRC.tracef(5,"CALC","enter calc fDetailedStatus inst|%s| name|%s| val|%s|" \
        % (propinst.sFullname,myvarname,propinst.value))
# detailedstatus
# non-recov = same as healthstate
# printerdetectederrorstate=low paper = stressed
# printerdetectederrorstate=low toner = stressed
# printerdetectederrorstate=outputnearfull = stressed
# printerdetectederrorstate= overdue preventive maint = stressed
# no additional for all other cases
    tempstat = fGetIntVal(propinst,myvarname,5)
    temperr = fGetIntVal(propinst,"hrPrinterDetectedErrorState",0)
    if tempstat & 16:           # if critical bit, 
        result = 4              # then critical failure.
    elif (tempstat & 7) == 3:   # if broken,
        result = 4              # then critical failure.
    elif temperr & 1:           # if low paper, 
        result = 2              # then stressed.
    elif temperr & 4:           # if low toner, 
        result = 2              # then stressed.
    elif temperr & 2048:        # if output near full, 
        result = 2              # then stressed.
    elif temperr & 16384:       # if overdue preventive maintenance, 
        result = 2              # then stressed.
    else:
        result = 1              # otherwise no additional information.
    return fCommonExit(propinst,myvarname,result,"fDetailedStatus")

# f O p e r a t i n g S t a t u s ( )
#
# Extract the OperatingStatus from a SubUnitStatus variable.  
def fOperatingStatus(propinst,myvarname):
    #myvarname = dStatusprop2var.mget(propinst.sFullname)
    TRC.tracef(5,"CALC","enter calc fOperatingStatus inst|%s| name|%s| val|%s|" \
        % (propinst.sFullname,myvarname,propinst.value))
# operatingstatus
# transitioning = transitioning
# unknown = unknown
# inservice otherwise
    temp = fGetIntVal(propinst,myvarname,5)
    if temp & 64:           # if transitioning, 
        result = 15         # then transitioning.
    elif (temp & 7) == 5:   # if unknown, 
        result = 0          # then unknown.
    else:
        result = 16         # otherwise in service.
    return fCommonExit(propinst,myvarname,result,"fOperatingStatus")

# f H e a l t h S t a t e ( )
#
# Extract the HealthState from a SubUnitStatus variable.  
def fHealthState(propinst,myvarname):
    #myvarname = dStatusprop2var.mget(propinst.sFullname)
    TRC.tracef(5,"CALC","enter calc fHealthState inst|%s| name|%s| val|%s|" \
        % (propinst.sFullname,myvarname,propinst.value))
    # critical bit = crit fail
    # broken = crit fail
    # non-crit bit = degraded
    temp = fGetIntVal(propinst,myvarname,5)
    if temp & 16:           # if critical bit, 
        result = 25         # then critical failure
    elif (temp & 7) == 3:   # if broken,
        result = 25         # then critical failure
    elif temp & 8:          # if non-critical bit, 
        result = 10         # then degraded
    elif (temp & 7 == 5):   # if unknown, 
        result = 0          # then unknown
    else:
        result = 5      # otherwise, must be okay
    return fCommonExit(propinst,myvarname,result,"fHealthState")


#=========== H e l p e r   F u n c t i o n s

# Helper functions for all of these calculation functions.

# f G e t I n t V a l ( )
#
# Return the integer value of the named input variable in the
# property's dictionary.  If the value doesn't convert nicely
# to an integer, return a default value.  
def fGetIntVal(mypropinst,myvarname,mydefaultval):
    stemp = eval("mypropinst.mget(\"%s\")"%myvarname)
    TRC.trace(5,"enter fGetIntVal prop|%s| var|%s| val|%s| default|%s|"\
        %(mypropinst.sFullname,myvarname,stemp,mydefaultval))
    try:
        temp = int(stemp)
    except (ValueError,TypeError):
        temp = mydefaultval # if not an integer value, make it read "unknown"
    return temp

# f G e t N i c e V a l ( )
#
# Return the integer value of the named input variable in the
# property's dictionary.  If the value doesn't convert nicely
# to a nice, polite, nonnegative integer, return a default value.  
def fGetNiceVal(mypropinst,myvarname,mydefaultval):
    stemp = eval("mypropinst.mget(\"%s\")"%myvarname)
    TRC.trace(5,"enter fGetNiceVal prop|%s| var|%s| val|%s| default|%s|"\
        %(mypropinst.sFullname,myvarname,stemp,mydefaultval))
    try:
        temp = int(stemp)           # If it is an integer value, 
        if temp < 0:                # make it a nonnegative value;
            temp = 0
    except (ValueError,TypeError):
        if stemp is None:
            temp = mydefaultval     # if not there, use default;
        else:
            temp = mydefaultval     # if not an integer value, make it read "unknown"
    return temp

# f G e t S t r V a l ( )
#
# Return the string value of the named input variable in the
# property's dictionary.  If the value is absent, 
# return a default value.  
def fGetStrVal(mypropinst,myvarname,mydefaultval):
    stemp = eval("mypropinst.mget(\"%s\")"%myvarname)
    if stemp == None \
    or stemp.lower().find("error")>=0 \
    or stemp.lower().find("not found")>=0:
        temp = mydefaultval # if not a real value, make it the default value
    else:
        temp = stemp
    return temp

# f G e t V a l 
#
# Return the value of any old thing, regardless of type.
def fGetVal(mypropinst,myvarname,mydefaultval):
    stemp = eval("mypropinst.mget(\"%s\")"%myvarname)
    if stemp == None:
        temp = mydefaultval # if not a real value, make it the default value
    else:
        temp = stemp
    return temp

# f C o m m o n E x i t ( )
# 
# Shorten all of these routines by four or five lines.  
# Store the result back into the property, log the output, 
# and return the value.  
def fCommonExit(mypropinst,myvarname,myresult,myfunctionname):
    mypropinst.value = myresult
    mypropinst.mset("value",mypropinst.value)
    TRC.tracef(2,"CALC","exit  commonexit calc %s inst|%s| name|%s| val|%s|" \
        % (myfunctionname,mypropinst.sFullname,myvarname,mypropinst.value))
    return mypropinst.value

# f M a k e B o o l e a n ( )
#
# Regularize any old integer values into proper CIM booleans.
def fMakeBoolean(myvalue):
    if int(myvalue) > 0:
        result = "TRUE"
    else:
        result = "FALSE"
    return result

# f O t h e r C l a s s _ G e t P r o p 
#
# Get me the value of some property from some other class.  
# Return the value of this property from this class in this row.
# 
# It is often necessary to retrieve values from other subsystems
# in order to calculate a property value.  This function packages 
# up all the calls needed to instantiate and evaluate the property.
def fOtherClass_GetProp(mypropname,myclassname,myrowid):
    cGetVal = fOtherClass_GetPropRef(mypropname,myclassname,myrowid)
    sVal = cGetVal.value
    TRC.tracef(2,"CALC","exit  calcother class|%s| prop|%s| row|%s| val|%s|"\
        %(myclassname,mypropname,myrowid,sVal))
    return sVal

# f O t h e r C l a s s _ G e t P r o p R e f 
#
# Get me the value of some property from some other class.  
# Return a ref to this property from this class in this row.
# 
# It is often necessary to retrieve values from other subsystems
# in order to calculate a property value.  This function packages 
# up all the calls needed to instantiate and evaluate the property.
def fOtherClass_GetPropRef(mypropname,myclassname,myrowid):
    from cimprop    import CCimProp
    cGetVal = CCimProp(mypropname,myrowid,None,myclassname)
    cGetVal.clearValue()
    cGetVal.getAllVars()
    cGetVal.generateValue()
    return cGetVal


#=========== C o m m o n   P i e c e s P a r t s   f u n c t i o n s

# f I n s t a n c e I D ( )
#
# Calculate the InstanceID property of any instance.  
# Include printer name, sysname, device description, 
# classname, and row number.  
def fInstanceID(propinst,myvarname):
    temp = fGetStrVal(propinst,myvarname,"nodata")
    tempsys = fGetStrVal(propinst,"sysName","nodata")
    if tempsys == "":
        tempsys = "empty_sysName"
    tempdesc = fGetStrVal(propinst,"hrDeviceDescr","nodata")
    propinst.value = "Dell:%s:%s:%s:%s:%s"\
        %(datastore.cPrinter.Name,tempsys,tempdesc,propinst.sClassname,temp)
    propinst.mset("InstanceID",propinst.value)
    TRC.tracef(2,"CALC","exit  calc fInstanceID class|%s| var|%s| val|%s|"\
        %(propinst.sFullname,temp,propinst.value))
    return propinst.value

# f N a m e ( )
# 
# Concatenate sysname, classname, and whatever arg is fed in.  
def fName(propinst,myvarname):
    tempclass = propinst.sClassname
    tempnum = fGetStrVal(propinst,myvarname,"nodata")
    tempsys = fGetStrVal(propinst,"sysName","nodata")
    if tempsys == "":
        tempsys = datastore.cPrinter.Name
    result = "%s %s %s" % (tempsys,tempclass,tempnum)
    return fCommonExit(propinst,myvarname,result,"fName")

    propinst.value = tempname
    propinst.mset("value",propinst.value)
    TRC.tracef(2,"CALC","exit  calc fName class|%s| val|%s|"\
        %(propinst.sFullname,propinst.value))
    return propinst.value

# f R e m a i n i n g C a p a c i t y ( )
#
# General routine to calculate capacity that might have those 
# funny out of band values -1, -2, -3, like that.  
# Coerce the value to be nonnegative or null.
def fRemainingCapacity(propinst,myvarname=""):
    temp = fGetIntVal(propinst,myvarname,0)
    if temp >= 0:
        result = temp
    elif temp == -3:    # If at least one
        result = 1      # then say that there's one.
    else:
        result = None
    propinst.value = result
    propinst.mset("value",propinst.value)
    TRC.tracef(2,"CALC","exit  calc fCapacity inst|%s| name|%s| val|%s|" \
        % (propinst.sFullname,myvarname,propinst.value))
    return propinst.value

# f B a s i s ( )
#
# Calculate the XxxBasis property attached to one of the Xxx properties
# that is really a counter but uses the funny out-of-band negative values
# for unknown, other, at-least-one, and so forth.
# A client ap should read the basis first to see if the value in the 
# Capacity or Level property is meaningful.  
def fBasis(propinst,myvarname):
    stemp = eval("propinst.mget(\"%s\")"%myvarname)
    try:
        temp = int(stemp)
    except (ValueError,TypeError):
        temp = -2       # if not an integer value, make it read "unknown"
    if temp ==  -3:     # at least one unit remains
        result = 4
    elif temp == -2:    # unknown
        result = 2
    elif temp == -1:    # other, unlimited
        result = 1
    else:               # otherwise, actual
        result = 3
    propinst.value = result
    propinst.mset("value",propinst.value)
    TRC.tracef(2,"CALC","exit  calc fBasis inst|%s| name|%s| val|%s|" \
        % (propinst.sFullname,myvarname,propinst.value))
    return propinst.value

# f I s D e f a u l t ( )
#
# Is this subunit instance the default one among its set?
def fIsDefault(propinst,myvarname):
    stemp = fGetStrVal(propinst,myvarname,"")
    if stemp.strip() == propinst.sRow.strip():
        result = 1
    else:
        result = 0
    sresult = fMakeBoolean(result)
    return fCommonExit(propinst,myvarname,sresult,"fIsDefault")
    


#=========== P r o p e r t y - s p e c i f i c   m a p p i n g s

# The functions used to compute property values from SNMP variables.
# In general the names are just property names, 



#===== A l e r t L o g 

#===== A l e r t R e c o r d 

# f A l e r t R e c o r d _ C o d e N a m e ( )
#
def fAlertRecord_CodeName(propinst,myvarname):
    stemp = fGetStrVal(propinst,myvarname,"Absent") 
    result = stemp
    result = "NYI: replace this with the IANA registered code name"
    return fCommonExit(propinst,myvarname,result,"fAlertRecord_CodeName")

# f A l e r t R e c o r d _ C o m p o n e n t C l a s s N a m e ( )
#
# In SNMP, subunits are numbered.  Convert the numbers to CIM
# classnames.  
def fAlertRecord_ComponentClassName(propinst,myvarname=""):
    temp = fGetIntVal(propinst,myvarname,0)
    dClassnames = dict( {   5 : "CIM_Printer", 6 : "CIM_PrintInterlock", 7 : "CIM_Printer", \
                    8 : "CIM_PrintInputTray", 9 : "CIM_PrintOutputTray", \
                    10 : "CIM_PrintMarker", 11 : "CIM_PrintSupply", \
                    12 : "CIM_PrintSupply", 13 : "CIM_PrintMediaPath", \
                    14 : "CIM_PrintChannel", 15 : "CIM_PrintInterpreter", \
                    16 : "CIM_PrintPrinter", 17 : "CIM_Printer", \
                    19 : "CIM_PrintAlertRecord", \
                    30 : "CIM_PrintFinisher", 31 : "CIM_PrintFinisher", 
                    32 : "CIM_PrintSupply", 33 : "CIM_PrintFinisher" } )
    if temp in dClassnames:
        result = dClassnames[temp]
    else:
        result = "Error: prtAlertGroup does not correspond to a class"
    return fCommonExit(propinst,myvarname,result,"fAlertRecord_ComponentClassName")

# f A l e r t R e c o r d _ C o m p o n e n t E l e m e n t N a m e ( )
# 
# This is the first tricky one.  We need the Name property of the
# instance of the subunit causing the alert.  This property comes from 
# a different subunit (table) than the alert table in which it appears.
# So first we have to get the Classname corresponding to the subunit
# of the alert cause, then we have to get the Name property from the 
# correct instance of that class.  In both cases, we instantiate and 
# evaluate the property out of the context of its entire instance.  
# Amazingly, that works.  We could instantiate the entire class 
# instance, too; just takes longer.  
def fAlertRecord_ComponentElementName(propinst,myvarname):
    # If you're going to create and access other properties, you have to 
    # import the property class.  
    from cimprop    import CCimProp
    # Get the instance number of the bad subunit.
    iInstnum = fGetIntVal(propinst,"prtAlertGroupIndex",1)
    # Get the classname of the offending unit.
    cGetClassname = CCimProp("ComponentClassName",propinst.sRow,None,"CIM_PrintAlertRecord")
    cGetClassname.clearValue()
    cGetClassname.getAllVars()
    cGetClassname.generateValue()
    sClassname = cGetClassname.getValue()
    # Now get the name property of that instance of the subunit. 
    if iInstnum > 0: 
        cGetElementname = CCimProp("Name",iInstnum,None,sClassname)
        cGetElementname.clearValue()
        cGetElementname.getAllVars()
        cGetElementname.generateValue()
        sElementname = cGetElementname.getValue()
        result = sElementname
    else:
        result = "element not specified"
    return fCommonExit(propinst,myvarname,result,"fAlertRecord_ComponentElementName")

# f A l e r t R e c o r d _ C r e a t i o n T i m e S t a m p ( )
#
def fAlertRecord_CreationTimeStamp(propinst,myvarname):
    iAlerttime = fGetIntVal(propinst,"prtAlertTime",0)
    iUptime = fGetIntVal(propinst,"sysUpTime",0)
    sUptime = fGetStrVal(propinst,"sysUpTime","0:0:0:0.0")
    nTicks = 0
    TRC.trace(5,"proc  AlertCreationTime before itime|%s| stime|%s| nticks|%s|"\
        %(iUptime,sUptime,nTicks))
    # Exception for the SNMP MIBdump files that format sysUptime as 
    # days:hours:minutes:seconds.hundredths.  Will never be used
    # with real SNMP output, where uptime is always an integer.  
    if iUptime == 0:    
        regUptime = "(\d+):(\d+):(\d+):(\d+)\.(\d+)"
        mMatchObj = re.search(regUptime,sUptime)
        if mMatchObj:
            (nDays,nHours,nMinutes,nSeconds,nHundredths) = \
                (int(mMatchObj.group(1)),int(mMatchObj.group(2)),int(mMatchObj.group(3)),int(mMatchObj.group(4)),int(mMatchObj.group(5)))
            nTicks = (((((nDays*24)+nHours)*60)+nMinutes)*60+nSeconds)*100 + nHundredths
        else:
            nTicks = 0
        iUptime = int(nTicks)
    # Exception for the SNMP MIBdump files that format sysUptime as 
    # timeticks in parens and then other stuff.  Will never be used
    # with real SNMP output, where uptime is always an integer.  
    if iUptime == 0:    
        regUptime = "^\((\d+)\)"
        mMatchObj = re.search(regUptime,sUptime)
        if mMatchObj:
            nTicks = int(mMatchObj.group(1))
        else:
            nTicks = 0
        iUptime = int(nTicks)
    TRC.trace(5,"proc  AlertCreationTime after itime|%s| stime|%s| nticks|%s|"\
        %(iUptime,sUptime,nTicks))
    dSecsBeforeNow = (abs(iUptime - iAlerttime)*1.0)/100.0
    dHdthsLeft = (abs(iUptime - iAlerttime)*1.0) % 100.
    dMinsBeforeNow = dSecsBeforeNow / 60
    dSecsLeft = dSecsBeforeNow % 60
    dHrsBeforeNow = dMinsBeforeNow / 60
    dMinsLeft = dMinsBeforeNow % 60
    dDaysLeft = dHrsBeforeNow / 24
    dHrsLeft = dHrsBeforeNow % 24
    strTimeBeforeNow = "%d:%02d:%02d:%02d.%02d"\
        %(dDaysLeft,dHrsLeft,dMinsLeft,dSecsLeft,dHdthsLeft)
    dNow = time.time()
    dThen = ( dNow - dSecsBeforeNow )
    result = time.strftime( "%Y-%m-%dT%H:%M:%S",time.gmtime(dThen) )
    result += "  %s ago" % (strTimeBeforeNow)
    return fCommonExit(propinst,myvarname,result,"fAlertRecord_CreationTimeStamp")

# f A l e r t R e c o r d _ R e c o r d F o r m a t ( ) 
# 
# The record format for alert data is fixed.  The MOF says to include
# all the numeric items in the prtAlert record.
def fAlertRecord_RecordFormat(propinst,myvarname):
    result = ';uint32 prtAlertIndex;uint32 prtAlertSeverityLevel;uint32 prtAlertTrainingLevel;uint32 prtAlertGroup;uint32 prtAlertGroupIndex;uint32 prtAlertLocation;uint32 prtAlertCode;uint32 prtAlertTime;'
    return fCommonExit(propinst,myvarname,result,"fAlertRecord_RecordFormat")

# f A l e r t R e c o r d _ R e c o r d D a t a ( )
# 
# List all the numeric items of the alert record in the order
# specified in the RecordFormat.  
def fAlertRecord_RecordData(propinst,myvarname):
    propinst.value = "this will be the record data someday"
    sname = "prtAlertIndex"
    temp1 = fGetNiceVal(propinst,sname,0)
    sname = "prtAlertSeverityLevel"
    temp2 = fGetNiceVal(propinst,sname,0)
    sname = "prtAlertTrainingLevel"
    temp3 = fGetNiceVal(propinst,sname,0)
    sname = "prtAlertGroup"
    temp4 = fGetNiceVal(propinst,sname,0)
    sname = "prtAlertGroupIndex"
    temp5 = fGetNiceVal(propinst,sname,0)
    sname = "prtAlertLocation"
    temp6 = fGetNiceVal(propinst,sname,0)
    sname = "prtAlertCode"
    temp7 = fGetNiceVal(propinst,sname,0)
    sname = "prtAlertTime"
    temp8 = fGetNiceVal(propinst,sname,0)
    result = ";%s;%s;%s;%s;%s;%s;%s;%s;" % (temp1,temp2,temp3,temp4,temp5,temp6,temp7,temp8)
    return fCommonExit(propinst,myvarname,result,"fAlertRecord_RecordData")

# fAlertRecord_LogInstanceID
# 
def fAlertRecord_LogInstanceID(propinst,myvarname):
    spropval = fOtherClass_GetProp("InstanceID","CIM_RecordLog",999)
    sresult = spropval
    return fCommonExit(propinst,myvarname,sresult,"fAlertRecord_LogInstanceID")


#===== C h a n n e l 


#===== I n p u t T r a y

# f I n p u t T r a y _ I n s t a n c e I D ( )
#
def fInputTray_InstanceID(propinst,myvar=""):
    temp = propinst.mget("prtInputIndex")
    propinst.value = "Dell:%s:InputTray:%s"%(datastore.cPrinter.Name,temp)
    TRC.tracef(2,"CALC","exit  calc fInputInstanceID class|%s| val|%s|"\
        %(propinst.sFullname,propinst.value))
    return propinst.value

# f I n p u t T r a y _ E l e m e n t N a m e ( )
#
def fInputTray_ElementName(propinst,myvar=""):
    temp = propinst.mget("prtInputName")
    if temp == "":
        tempnum = propinst.mget("prtInputIndex")
        propinst.value = "InputTray %s"%(tempnum)
    TRC.tracef(2,"CALC","exit  calc fInputName class|%s| val|%s|"\
        %(propinst.sFullname,propinst.value))
    return propinst.value

# f I n p u t T r a y _ I s D e f a u l t ( )
#
def fInputTray_IsDefault(propinst,myvarname):
    if propinst.mget("prtInputDefaultIndex") == propinst.sRow:
        result = 1
    else:
        result = 0
    sresult = fMakeBoolean(result)
    return fCommonExit(propinst,myvarname,sresult,"fInputTray_IsDefault")
    
# f I n p u t T r a y _ C u r r e n t L e v e l ( )
#
def fInputTray_CurrentLevel(propinst,myvar=""):
    temp = int( propinst.mget("prtInputCurrentLevel") )
    if temp >= 0:
        propinst.value = temp
    elif temp == -3:
        propinst.value = 1
    else:
        propinst.value = None
    TRC.tracef(2,"CALC","exit  calc fInputCurrentLevel class|%s| val|%s|"\
        %(propinst.sFullname,propinst.value))
    return propinst.value

# f I n p u t T r a y _ M a x C a p a c i t y ( )
#
def fInputTray_MaxCapacity(propinst,myvar=""):
    temp = int( propinst.mget("prtInputMaxCapacity") )
    if temp >= 0:
        propinst.value = temp
    else:
        propinst.value = None
    TRC.tracef(2,"CALC","exit  calc fInputMaxCapacity class|%s| val|%s|"\
        %(propinst.sFullname,propinst.value))
    return propinst.value

# f I n p u t T r a y _ M e d i a S i z e N a m e 
# 
def fInputTray_MediaSizeName(propinst,myvarname):
    dimunit = fGetIntVal(propinst,myvarname,0)
    dim1 = fGetIntVal(propinst,"prtInputMediaDimFeedDirChosen",0)
    dim2 = fGetIntVal(propinst,"prtInputMediaDimXFeedDirChosen",0)
    if dim1 == 0:
        dim1 = fGetIntVal(propinst,"prtInputMediaDimFeedDirDeclared",0)
    if dim2 == 0:
        dim2 = fGetIntVal(propinst,"prtInputMediaDimXFeedDirDeclared",0)
    if dim1 > dim2:
        (dimlong,dimshort) = (dim1,dim2)
    else:
        (dimlong,dimshort) = (dim2,dim1)
    TRC.trace(5,"enter fInputTray_MediaSizeName unit|%s| short|%s| long|%s|"\
        %(dimunit,dimshort,dimlong))
    sresult = fGetPaperSize(dimunit,dimshort,dimlong)
    if dimshort <= 0 or dimlong <= 0:
        sresult = "((zero: not specified))"
    return fCommonExit(propinst,myvarname,sresult,"fInputTray_MediaSizeName")

# function f G e t P a p e r S i z e ( ) 
# 
# Used by a number of properties.
# For the moment, cover the common sizes that we've seen lately.  
# 
# The old approach of exact comparison just wasn't working for us:
# every new printer seems to have a different impression of how long
# 8-1/2 inches is in micrometers.  And the same for other lengths.  
# This is a somewhat more sophisticated approach to matching
# sizes, such as some nominal size + or - say 2%.  The set of nominal 
# values should include all the standard values for paper, envelopes, 
# North American, European, Japanese.  Just tabulate the tolerance values
# and search the table any old way.  Linear search suffices for the small
# number of items we will have.  Each interval maps back to its nominal 
# length; a pair of nominal lengths define a sheet size.  
# If any measurement does not fall into a known interval, or if the pair
# of lengths does not match any known size, then declare it to be custom.  
def fGetPaperSize(mydimunit,mydimshort,mydimlong):
    sresult = None
    strSize = "unit-not-specified"
    if mydimunit == 3:    # ten thousandths of inches
        strSize = "%.1fx%.1fin"%(mydimshort/10000.0,mydimlong/10000.0)
        if mydimlong == 110000 and mydimshort == 85000:
            sresult = "na_letter_8.5x11in"
        if mydimlong == 140000 and mydimshort == 85000:
            sresult = "na_legal_8.5x14in"
        if mydimlong == 170000 and mydimshort == 110000:
            sresult = "na_ledger_11x17in"
        if mydimlong == 95000 and mydimshort == 41259:
            sresult = "na-number-10-envelope"
    elif mydimunit == 4:  # micrometers
        strSize = "%.1fx%.1fmm"%(mydimshort/1000.0,mydimlong/1000.0)
        try:
            iShort = int(mydimshort)
            iLong = int(mydimlong)
        except (ValueError,TypeError):
            iShort = iLong = 0
        sShortname = sLongname = None
        for sSizename in dPapersize2limits.keys():
            (iLimitLo,iLimitHi) = dPapersize2limits[sSizename]
            if iShort >= iLimitLo and iShort <= iLimitHi:
                sShortname = sSizename
            if iLong >= iLimitLo and iLong <= iLimitHi:
                sLongname = sSizename
        TRC.trace(5,"proc  fGetPaperSize short|%d| name|%s| long|%d| name|%s|"%\
            (iShort,sShortname,iLong,sLongname))
        if sShortname == "eightandahalf" and sLongname == "eleven":
            sresult = "na_letter_8.5x11in"
        if sShortname == "eightandahalf" and sLongname == "fourteen":
            sresult = "na_legal_8.5x14in"
        if sShortname == "eleven" and sLongname == "seventeen":
            sresult = "na_ledger_11x17in"
        if sShortname == "fourandaneighth" and sLongname == "nineandahalf":
            sresult = "na-number-10-envelope"
        if sShortname == "isoa4width" and sLongname == "isoa4length":
            sresult = "iso_a4_210x297mm"
        if sShortname == "isoa4length" and sLongname == "isoa3length":
            sresult = "iso_a3_297x420mm"
        TRC.trace(3,"proc  fGetPaperSize short|%d| long|%d| result|%s|"%\
            (iShort,iLong,sresult))
        '''
        if mydimlong == 279400 and mydimshort == 215900:
            sresult = "na_letter_8.5x11in"
        if mydimlong == 279000 and mydimshort == 215900:
            sresult = "na_letter_8.5x11in"
        if mydimlong == 279000 and mydimshort == 216000:
            sresult = "na_letter_8.5x11in"
        if mydimlong == 355600 and mydimshort == 215900:
            sresult = "na_legal_8.5x14in"
        if mydimlong == 356000 and mydimshort == 216000:
            sresult = "na_legal_8.5x14in"
        if mydimlong == 431800 and mydimshort == 279400:
            sresult = "na_ledger_11x17in"
        if mydimlong == 432000 and mydimshort == 279000:
            sresult = "na_ledger_11x17in"
        elif mydimlong == 297000 and mydimshort == 210000:
            sresult = "iso_a4_210x297mm"
        '''
    if sresult == None:
        sresult = "custom_" + strSize
    return sresult


#===== I n t e r l o c k 

# f I n t e r l o c k _ S t a t u s ( )
#
# A leftover from the old days that should be replaced.  
def fInterlock_Status(propinst,myvarname):
    result = fGetIntVal(propinst,"prtCoverStatus",0) 
    return fCommonExit(propinst,myvarname,result,"fInterlock_Status")

# f I n t e r l o c k _ N a m e ( )
#
def fInterlock_Name(propinst,myvar=""):
    temp = propinst.sClassname
    tempnum = propinst.mget("prtCoverIndex")
    tempname = temp
    if temp == "" or temp == None:
        tempname = "Interlock"
    else:
        tempname = temp
    propinst.value = "%s %s"%(tempname,tempnum)
    TRC.tracef(2,"CALC","exit  calc fInterlockName class|%s| val|%s|"\
        %(propinst.sFullname,propinst.value))
    return propinst.value

# f I n t e r l o c k _ E l e m e n t N a m e ( )
#
def fInterlock_ElementName(propinst,myvavname):
    temp = propinst.mget("prtCoverDescription")
    tempnum = propinst.mget("prtCoverIndex")
    tempname = temp
    if temp == "" or temp == None:
        tempname = "Interlock"
    else:
        tempname = temp
    propinst.value = "%s %s"%(tempname,tempnum)
    TRC.tracef(2,"CALC","exit  calc fInterlockElementName class|%s| val|%s|"\
        %(propinst.sFullname,propinst.value))
    return propinst.value


#===== I n t e r p r e t e r 


#===== M a r k e r


#===== M e d i a P a t h 

# f M e d i a P a t h _ M a x S i z e 
#
def fMediaPath_MaxSize(propinst,myvarname):
    dimunit = fGetIntVal(propinst,myvarname,0)
    dim1 = fGetIntVal(propinst,"prtMediaPathMaxMediaFeedDir",0)
    dim2 = fGetIntVal(propinst,"prtMediaPathMaxMediaXFeedDir",0)
    if dim1 > dim2:
        (dimlong,dimshort) = (dim1,dim2)
    else:
        (dimlong,dimshort) = (dim2,dim1)
    TRC.trace(5,"enter fMediaPath_MaxMediaSizeName unit|%s| short|%s| long|%s|"\
        %(dimunit,dimshort,dimlong))
    sresult = fGetPaperSize(dimunit,dimshort,dimlong)
    if dimshort <= 0 or dimlong <= 0:
        sresult = "((zero: not specified))"
    return fCommonExit(propinst,myvarname,sresult,"fInputTray_MediaSizeName")

# f M e d i a P a t h _ M i n S i z e 
# 
def fMediaPath_MinSize(propinst,myvarname):
    dimunit = fGetIntVal(propinst,myvarname,0)
    dim1 = fGetIntVal(propinst,"prtMediaPathMinMediaFeedDir",0)
    dim2 = fGetIntVal(propinst,"prtMediaPathMinMediaXFeedDir",0)
    if dim1 > dim2:
        (dimlong,dimshort) = (dim1,dim2)
    else:
        (dimlong,dimshort) = (dim2,dim1)
    TRC.trace(5,"enter fMediaPath_MinMediaSizeName unit|%s| short|%s| long|%s|"\
        %(dimunit,dimshort,dimlong))
    sresult = fGetPaperSize(dimunit,dimshort,dimlong)
    if dimshort <= 0 or dimlong <= 0:
        sresult = "((zero: not specified))"
    return fCommonExit(propinst,myvarname,sresult,"fInputTray_MediaSizeName")


#===== O u t p u t T r a y 

# f O u t p u t T r a y _ I s D e f a u l t ( )
#
def fOutputTray_IsDefault(propinst,myvarname):
    if propinst.mget("prtOutputDefaultIndex") == propinst.sRow:
        result = 1
    else:
        result = 0
    return fCommonExit(propinst,myvarname,result,"fOutputTray_IsDefault")
    
# f O u t p u t T r a y _ O f f l i n e S t a t u s ( )
#
def fOutputTray_OfflineStatus(propinst,myvarname):
    tempval = eval("propinst.mget(\"%s\")"%myvarname) 
    if tempval != None:
        temp = int(tempval)
        temp = temp & 32
        result = temp
    else:
        result = 0
    propinst.value = result
    propinst.mset("value",propinst.value)
    TRC.tracef(2,"CALC","exit  calc fOutputTray_OfflineStatus inst|%s| name|%s| val|%s|" \
        % (propinst.sFullname,myvarname,propinst.value))
    return propinst.value

# f O u t p u t T r a y _ R e m a i n i n g C a p a c i t y ( )
#
def fOutputTray_RemainingCapacity(propinst,myvarname):
    temp = fGetIntVal(propinst,myvarname,-1)
    if temp >= 0:
        result = temp
    else:
        result = None
    return fCommonExit(propinst,myvarname,result,"fOutputTray_RemainingCapacity")


#===== P r i n t e r 

# f P r i n t e r _ P r i n t e r K e y ( )
#
def fPrinter_PrinterKey(propinst,myvarname):
    pass
    sSystemCreationClassName = fOtherClass_GetProp("SystemCreationClassName","CIM_Printer",1)
    sCreationClassName = fOtherClass_GetProp("CreationClassName","CIM_Printer",1)
    sSystemName = fOtherClass_GetProp("SystemName","CIM_Printer",1)
    sName = fOtherClass_GetProp("Name","CIM_Printer",1)
    sresult = "Classname=\"%s\";SystemCreationClassName=\"%s\";SystemName=\"%s\";CreationClassName=\"%s\";Name=\"%s\";"\
        %("CIM_Printer",sSystemCreationClassName,sSystemName,sCreationClassName,sName) 
    #sresult = datastore.cPrinter.lPrinterKey
    return fCommonExit(propinst,myvarname,sresult,"fPrinterKey")


def fPrinter_GetProperty(mypropname,myclassname,myrowid):
    from cimprop    import CCimProp
    cGetVal = CCimProp(mypropname,myrowid,None,myclassname)
    cGetVal.clearValue()
    cGetVal.getAllVars()
    cGetVal.generateValue()
    sVal = cGetVal.value
    return sVal

# f P r i n t e r _ C o n s o l e D i s a b l e d 
# 
def fPrinter_ConsoleDisabled(propinst,myvarname):
    sVal = fOtherClass_GetProp("PROPprtConsoleDisable","ElFako_Console",1)
    if sVal == 3:
        temp = 0
    elif sVal == 4:
        temp = 1
    else:
        temp = 0
    sresult = fMakeBoolean(temp)
    return fCommonExit(propinst,myvarname,sresult,"ConsoleDisabled")

# f P r i n t e r _ C o n s o l e D i s p l a y B u f f e r T e x t 
def fPrinter_ConsoleDisplayBufferText(propinst,myvarname):
    pass
    # Get the number of lines in the display.  
    temp = fGetIntVal(propinst,"prtConsoleNumberOfDisplayLines",1)
    from cimprop    import CCimProp
    # For each line, create a syntho property to hold the line, and get its value.
    for idxRow in range(1,temp+1):
        cGetVal = CCimProp("PROPprtConsoleDisplayBufferText",idxRow,None,"ElFako_ConsoleDisplay")
        cGetVal.clearValue()
        cGetVal.getAllVars()
        cGetVal.generateValue()
        sVal = cGetVal.value
        templist = propinst.mget("valuelist")
        templist.append(sVal)
        propinst.mset("valuelist",templist)
        propinst.valuelist.append(sVal)
    propinst.value = propinst.valuelist
    lresult = propinst.value
    return fCommonExit(propinst,myvarname,lresult,"ConsoleDisplayBufferText")

# f P r i n t e r _ C o n s o l e N a t u r a l L a n g u a g e 
# 
def fPrinter_ConsoleNaturalLanguage(propinst,myvarname):
    sVal = fOtherClass_GetProp("PROPprtConsoleLocalization","ElFako_Console",1)
    sLocalization = fOtherClass_GetProp("PROPprtLocalizationLanguage","ElFako_Localization",sVal)
    sresult = sLocalization
    return fCommonExit(propinst,myvarname,sresult,"ConsoleNaturalLanguage")

# f P r i n t e r _ C u r r e n t N a t u r a l L a n g u a g e 
# 
def fPrinter_CurrentNaturalLanguage(propinst,myvarname):
    sVal = fOtherClass_GetProp("PROPprtGeneralCurrentLocalization","ElFako_Console",1)
    sLocalization = fOtherClass_GetProp("PROPprtLocalizationLanguage","ElFako_Localization",sVal)
    sresult = sLocalization
    return fCommonExit(propinst,myvarname,sresult,"CurrentNaturalLanguage")

# f P r i n t e r _ C u r r e n t C h a r S e t 
# 
def fPrinter_CurrentCharSet(propinst,myvarname):
    sVal = fOtherClass_GetProp("PROPprtGeneralCurrentLocalization","ElFako_Console",1)
    sCharset = fOtherClass_GetProp("PROPprtLocalizationCharacterSet","ElFako_Localization",sVal)
    sresult = sCharset
    return fCommonExit(propinst,myvarname,sresult,"CurrentCharSet")

# f P r i n t e r _ N a t u r a l L a n g u a g e s S u p p o r t e d 
# 
def fPrinter_NaturalLanguagesSupported(propinst,myvarname):
    lKeys = dClassname2instkeys["ElFako_Localization"]
    TRC.tracef(3,"CALC","proc  fPrinter_NaturalLanguagesSupported keys|%s|"%(lKeys))
    lLangs = list()
    for sRow in lKeys:
        sLocalization = fOtherClass_GetProp("PROPprtLocalizationLanguage","ElFako_Localization",sRow)
        lLangs.append(sLocalization)
    lresult = lLangs
    propinst.listvalue = lLangs
    return fCommonExit(propinst,myvarname,lresult,"NaturalLanguagesSupported")

# f P r i n t e r _ C h a r S e t s S u p p o r t e d 
# 
def fPrinter_CharSetsSupported(propinst,myvarname):
    lKeys = dClassname2instkeys["ElFako_Localization"]
    TRC.tracef(3,"CALC","proc  fPrinter_CharSetsSupported keys|%s|"%(lKeys))
    lLangs = list()
    for sRow in lKeys:
        sCharset = fOtherClass_GetProp("PROPprtLocalizationCharacterSet","ElFako_Localization",sRow)
        if True:    #if sCharset not in lLangs:  # Store unique values only.  The spec is ambiguous.  
            lLangs.append(sCharset)
    lresult = lLangs
    propinst.listvalue = lLangs
    return fCommonExit(propinst,myvarname,lresult,"CharSetsSupported")

# f P r i n t e r _ O p e r a t i n g S t a t u s 
# stub
def fPrinter_OperatingStatus(propinst,myvarname):
    sresult = "(calculate rollup of OperatingStatus of subunits)"
    return fCommonExit(propinst,myvarname,sresult,"PrinterOperatingStatus")

# f P r i n t e r _ O p e r a t i o n a l S t a t u s 
# stub
def fPrinter_OperationalStatus(propinst,myvarname):
    sresult = "(calculate rollup of OperationalStatus of subunits)"
    return fCommonExit(propinst,myvarname,sresult,"PrinterOperationalStatus")

# f P r i n t e r _ E n a b l e d S t a t e 
# 
def fPrinter_EnabledState(propinst,myvarname):
    temp = fGetIntVal(propinst,myvarname,0)
    if temp == 2:       # if running
        status = 2      #  then enabled
    elif temp == 5:     # if down
        status = 3      #  then disabled
    elif temp == 1:     # if unknown
        status = 0      #  then unknown
    else:               # else
        status = 5      #  not applicable, the default
    return fCommonExit(propinst,myvarname,status,"PrinterEnabledState")

# f P r i n t e r _ L a n g u a g e s S u p p o r t e d 
# 
def fPrinter_LanguagesSupported(propinst,myvarname):
    lKeys = dClassname2instkeys["CIM_PrintInterpreter"]
    TRC.tracef(3,"CALC","proc  fPrinter_LanguagesSupported keys|%s|"%(lKeys))
    lInterps = list()
    for sRow in lKeys:
        sInterp = fOtherClass_GetProp("Description","CIM_PrintInterpreter",sRow)
        lInterps.append(sInterp)
    lresult = lInterps
    propinst.listvalue = lresult
    return fCommonExit(propinst,myvarname,lresult,"NaturalLanguagesSupported")

# f P r i n t e r _ P a p e r T y p e s A v a i l a b l e 
# 
def fPrinter_PaperTypesAvailable(propinst,myvarname):
    lKeys = dClassname2instkeys["CIM_PrintInputTray"]
    TRC.tracef(3,"CALC","proc  fPrinter_PaperTypesAvailable keys|%s|"%(lKeys))
    lMediaNames = list()
    for sRow in lKeys:
        sMediaName = fOtherClass_GetProp("MediaSizeName","CIM_PrintInputTray",sRow)
        lMediaNames.append(sMediaName)
    lresult = lMediaNames
    propinst.listvalue = lresult
    return fCommonExit(propinst,myvarname,lresult,"Printer_PaperTypesAvailable")

# f P r i n t e r _ D e f a u l t P a p e r T y p e 
# 
def fPrinter_DefaultPaperType(propinst,myvarname):
    itemp = fGetIntVal(propinst,"prtInputDefaultIndex",0)
    if itemp == 0:
        sresult = "((default not specified))"
    else:
        sresult = fOtherClass_GetProp("MediaSizeName","CIM_PrintInputTray",itemp)
    return fCommonExit(propinst,myvarname,sresult,"Printer_DefaultPaperType")


#===== R e c o r d L o g 
#
# The RecordLog class is a singleton, not table-related.  

# f R e c o r d L o g _ I n s t a n c e I D 
# 
def fRecordLog_InstanceID(propinst,myvarname):
    temp = 999          # hardly matters for a singleton
    tempsys = fGetStrVal(propinst,"sysName","nodata")
    if tempsys == "":
        tempsys = "empty_sysName"
    tempdesc = fGetStrVal(propinst,"hrDeviceDescr","nodata")
    sresult = "Dell:%s:%s:%s:%s:%s"\
        %(datastore.cPrinter.Name,tempsys,tempdesc,propinst.sClassname,temp)
    return fCommonExit(propinst,myvarname,sresult,"RecordLog_InstanceID")
    

#===== S u p p l y

# f S u p p l y _ M a x C a p a c i t y ( )
#
def fSupply_MaxCapacity(propinst,myvar=""):
    temp = int( propinst.mget("prtMarkerSuppliesMaxCapacity") )
    if temp >= 0:
        result = temp
    else:
        result = None
    return fCommonExit(propinst,myvar,result,"fSupply_MaxCapacity")

# f S u p p l y _ C o l o r a n t N a m e 
# 
# All the colorant properties are done strangely.  
# First, the values for the properties are all accessed thru the 
# pseudo-class ElFako.  The values all come from the prtMarkerColorant
# table that is indexed separately and therefore cannot be accessed
# directly from a Supply property.
# Second, the correct row index for the the colorant is given in
# a variable in the prtMarkerSupplies table.  Retrieve that first 
# and use it to instantiate the correct el fako colorant property.  
def fSupply_ColorantName(propinst,myvar=""):
    TRC.tracef(1,"CALC","enter fSupply_ColorantName class|%s| row|%s|"\
        %(propinst.sClassname,propinst.sRow))
    from cimprop    import CCimProp
    temp = fGetIntVal(propinst,"prtMarkerSuppliesColorantIndex",0)
    # If the supply is not a colorant associated with a marker, its name is not relevant.
    if temp != 0:
        cGetName = CCimProp("PROPprtMarkerColorantValue",temp,None,"ElFako_Colorant")
        cGetName.clearValue()
        cGetName.getAllVars()
        cGetName.generateValue()
        sName = cGetName.value
        TRC.tracef(3,"CALC","proc fSupply_ColorantName class|%s| row|%s| name|%s|"\
            %(propinst.sClassname,propinst.sRow,sName))
        if sName == 0 or sName == "" or sName is None:
            sresult = "((supply is not a colorant))"
        else:
            sresult = sName
    else:
        sresult = "((supply is not a colorant))"
    return fCommonExit(propinst,myvar,sresult,"fSupply_ColorantName")

# f S u p p l y _ C o l o r a n t R o l e 
# 
def fSupply_ColorantRole(propinst,myvar=""):
    from cimprop    import CCimProp
    temp = fGetIntVal(propinst,"prtMarkerSuppliesColorantIndex",0)
    cGetProp = CCimProp("PROPprtMarkerColorantRole",temp,None,"ElFako_Colorant")
    cGetProp.clearValue()
    cGetProp.getAllVars()
    cGetProp.generateValue()
    nProp = cGetProp.value
    if nProp > 0:
        result = nProp
    else:
        result = None       # not a colorant at all
    return fCommonExit(propinst,myvar,result,"ColorantRole")

# f S u p p l y _ C o l o r a n t T o n a l i t y 
# 
def fSupply_ColorantTonality(propinst,myvar=""):
    from cimprop    import CCimProp
    temp = fGetIntVal(propinst,"prtMarkerSuppliesColorantIndex",0)
    # If the supply is not a colorant associated with a marker, its tonality is not relevant.
    if temp != 0:
        cGetProp = CCimProp("PROPprtMarkerColorantTonality",temp,None,"ElFako_Colorant")
        cGetProp.clearValue()
        cGetProp.getAllVars()
        cGetProp.generateValue()
        sProp = cGetProp.value
        if sProp != 0:
            sresult = sProp
        else:
            sresult = None
    else:
        sresult = None
    return fCommonExit(propinst,myvar,sresult,"ColorantTonality")

#===== E l F a k o C o l o r a n t  class
# 
# This is synthetic class that does not appear in the CIM model as a
# separate entity.  The properties are all direct reflections of their
# underlying SNMP variables.  
# Since these variables are in a separate table, separately indexed
# in snmp, it is hard to access them directly from any other class
# that is tied to another table.  So expose them into the properties 
# of a pseudo-class.  
# Why didn't I do this with fDirect?  Well, getIntVal() and getStrVal()
# both handle error cases and return reasonable default values.  The 
# fDirect mapping would let an "OID not found" error show through.  
# If that is desired instead, then use fDirect().

def fElFakoColorant_Index(propinst,myvar=""):
    result = fGetIntVal(propinst,"prtMarkerColorantIndex",0)
    return fCommonExit(propinst,myvar,result,"ColorantIndex")
    
def fElFakoColorant_MarkerIndex(propinst,myvar=""):
    result = fGetIntVal(propinst,"prtMarkerColorantMarkerIndex",0)
    return fCommonExit(propinst,myvar,result,"ColorantMarkerIndex")
    
def fElFakoColorant_Role(propinst,myvar=""):
    result = fGetIntVal(propinst,"prtMarkerColorantRole",0)
    return fCommonExit(propinst,myvar,result,"ColorantRole")
    
def fElFakoColorant_Value(propinst,myvar=""):
    sresult = fGetStrVal(propinst,"prtMarkerColorantValue",0)
    return fCommonExit(propinst,myvar,sresult,"ColorantValue")

def fElFakoColorant_Tonality(propinst,myvar=""):
    result = fGetIntVal(propinst,"prtMarkerColorantTonality",0)
    return fCommonExit(propinst,myvar,result,"ColorantTonality")
    
#===== E l F a k o C o n s o l e  class


#=========== A s s o c i a t i o n   C l a s s   F u n c t i o n s

# f P r i n t e r K e y ( )
#
# The key for CIM_Printer is the concatenation of four properties.  
# This is the old-style CIM key; more recent classes use InstanceID.
def fPrinterKey(propinst,myvarname):
    sresult = "Classname=\"CIM_Printer\";"
    lAllKeys = datastore.cPrinter.lPrinterKey
    for lOneKey in lAllKeys:
        sresult += "%s=\"%s\";"%(lOneKey[0],lOneKey[1])
    return fCommonExit(propinst,myvarname,sresult,"fPrinterKey")


#===== P r i n t e r C o m p o n e n t 

# f M a n d a t o r y C o m p o n e n t K e y ( ) 
#
# For CIM_PrinterComponent, get the InstanceID of the mandatory dependent 
# (subunit) class instance that is required for this printer.  
# Find the right classname and instance number, then instantiate
# the InstanceID property for that instance, and evaluate it.  
# We can instantiate the whole instance if necessary; it just takes
# a little longer and wastes some memory, but may be necessary in 
# cases where properties are interdependent.  
def fMandatoryComponentKey(propinst,myvarname):
    sresult = "mandatory component key for |%s| goes here" \
        %(propinst.sFullname)
    from cimclass   import CCimClass
    from cimprop    import CCimProp
    cAssocParent = propinst.refOwnerinst
    cAssocInstObj = cAssocParent.oInstObj
    cAssocInstInfo = cAssocInstObj.getInstInfo()
    sMandatoryClassname = cAssocInstInfo[0][1]
    nMandatoryClassRowId = cAssocInstInfo[1][1]
    TRC.trace(5,"proc  fMandCompKey assoc|%s| class2|%s| rowid|%s|"\
        %(propinst.sClassname,sMandatoryClassname,nMandatoryClassRowId))
    '''
    # This version instantiates the entire class just to get a property.
    cClass = CCimClass(sMandatoryClassname,"SNMPRowId=%s"%(nMandatoryClassRowId),nMandatoryClassRowId)
    cClass.makeAllProps()
    cClass.clearAllValues()
    cClass.getAllVars()
    cClass.generateAllValues()
    cInstanceID = cClass.dProp2inst.mget("InstanceID")
    TRC.trace(5,"proc  fMandCompKey cclass|%s| cprop|%s|"\
        %(id(cClass),id(cInstanceID)))
    sInstanceID = cInstanceID.value
    '''
    # This version instantiates only the property needed, no parent class.
    cGetInstanceID = CCimProp("InstanceID",nMandatoryClassRowId,None,sMandatoryClassname)
    cGetInstanceID.clearValue()
    cGetInstanceID.getAllVars()
    cGetInstanceID.generateValue()
    sInstanceID = cGetInstanceID.value
    sresult ="Classname=\"%s\";InstanceID=\"%s\""%(sMandatoryClassname,sInstanceID)
    return fCommonExit(propinst,myvarname,sresult,"fMandatoryComponentKey")


#===== C o n c r e t e C o m p o n e n t 

# f O p t i o n a l C o m p o n e n t K e y ( ) 
# 
# For CIM_COncreteComponent, get the InstanceID of the optional dependent
# (subunit) class instance that is attached to this printer.  
# Same procedure as for mandatory key: find the classname and instance
# number, and evaluate the InstanceID property for (what would be) 
# that instance.  
def fOptionalComponentKey(propinst,myvarname):
    sresult = "optional component key for |%s| goes here" \
        %(propinst.sFullname)
    from cimclass   import CCimClass
    from cimprop    import CCimProp
    cAssocParent = propinst.refOwnerinst
    cAssocInstObj = cAssocParent.oInstObj
    cAssocInstInfo = cAssocInstObj.getInstInfo()
    sConcreteClassname = cAssocInstInfo[0][1]
    nConcreteClassRowId = cAssocInstInfo[1][1]
    TRC.trace(5,"proc  fMandCompKey assoc|%s| class2|%s| rowid|%s|"\
        %(propinst.sClassname,sConcreteClassname,nConcreteClassRowId))
    cGetInstanceID = CCimProp("InstanceID",nConcreteClassRowId,None,sConcreteClassname)
    cGetInstanceID.clearValue()
    cGetInstanceID.getAllVars()
    cGetInstanceID.generateValue()
    sInstanceID = cGetInstanceID.value
    sresult ="Classname=\"%s\";InstanceID=\"%s\""%(sConcreteClassname,sInstanceID)
    return fCommonExit(propinst,myvarname,sresult,"fConcreteComponentKey")


#===== U s e O f L o g 

# f L o g K e y 
# 
def fLogKey(propinst,myvarname):
    spropval = fOtherClass_GetProp("InstanceID","CIM_RecordLog",999)
    sresult = "Classname=\"%s\";InstanceID=\"%s\"" \
                %("CIM_RecordLog",spropval)
    return fCommonExit(propinst,myvarname,sresult,"LogManagesRecord_AlertKey")


#===== L o g M a n a g e s R e c o r d 

# f A l e r t R e c o r d K e y 
# 
def fAlertRecordKey(propinst,myvarname):
    #temp = fOtherClass_GetProp("SNMPRowId","CIM_PrintAlertRecord",OOPSIES)
    (class1,class2) = fGetVal(propinst,"lRelatedClassname",[])
    (rowid1,rowid2) = fGetVal(propinst,"lSNMPRowId",[])
    if class1 == "CIM_PrintAlertRecord":
        temp = rowid1
    else:
        temp = rowid2
    spropval = fOtherClass_GetProp("InstanceID","CIM_PrintAlertRecord",temp)
    sresult = "Classname=\"%s\";InstanceID=\"%s\"" \
                %("CIM_PrintAlertRecord",spropval)
    return fCommonExit(propinst,myvarname,sresult,"LogManagesRecord_AlertKey")


#===== A s s o c i a t e d P r i n t I n t e r p r e t e r 

'''
The method used to produce all the instances of AssociatedPrintInterpreter is 
not perfect, and certainly not pretty, but works with the following caveat.

- If the channel uses different interpreters for job control and page description, 
this will produce duplicate entries for the association.  The duplicates will have 
to be filtered out by the mechanism that delivers the instances back to the 
caller.  Alternatively, we can insert an illegal value into the InterpreterType
property, and that might be easier to filter out at the last minute.  See the 
comments in the code about where to insert the illegal value.  -1 is my favorite.

This funny mechanism must be used in order to get the two separate instances
required in case the channel uses the same interpreter for both functions, 
which is often the case.  

Later: The InterpreterType function now deletes the instance if it thinks
it if about to create a duplicate instance.  It still isn't pretty, but it works.
'''

# f I n t e r p r e t e r K e y 
# 
def fInterpreterKey(propinst,myvarname):
    TRC.trace(1,"enter fInterpreterKey inst|%s| propname|%s|"\
        %(id(propinst),propinst.sFullname))
    from cimclass   import CCimClass
    from cimprop    import CCimProp
    sresult = "component key for |%s| goes here" \
        %(propinst.sFullname)
    cAssocParent = propinst.refOwnerinst        # class parent of property
    cAssocInstObj = cAssocParent.oInstObj       # CAssocInst instance for assoc class
    cAssocInstInfo = cAssocInstObj.getInstInfo()# list of both classnames, both rowids
    # Not sure what order the ini file classes were in, so look
    # for the right classname here.  Get classname and rowid.
    if cAssocInstInfo[0][0] == "CIM_PrintInterpreter":
        sConcreteClassname = cAssocInstInfo[0][0]
        nConcreteClassRowId = cAssocInstInfo[1][0]
    elif cAssocInstInfo[0][1] == "CIM_PrintInterpreter":
        sConcreteClassname = cAssocInstInfo[0][1]
        nConcreteClassRowId = cAssocInstInfo[1][1]
    else:
        TRC.trace(1,"error fInterpreterKey no Interpreter class in group|%s|"\
            %(cAssocInstInfo))
        return None
    TRC.trace(5,"proc  fInterpreterKey assoc|%s| class2|%s| rowid|%s|"\
        %(propinst.sClassname,sConcreteClassname,nConcreteClassRowId))
    sRow = nConcreteClassRowId
    # Get the InstanceID property from the right classname
    spropval = fOtherClass_GetProp("InstanceID",sConcreteClassname,sRow)
    sresult = spropval
    return fCommonExit(propinst,myvarname,sresult,"fInterpreterKey")

# f C h a n n e l K e y 
# 
def fChannelKey(propinst,myvarname):
    TRC.trace(1,"enter fChannelKey inst|%s| propname|%s|"\
        %(id(propinst),propinst.sFullname))
    from cimclass   import CCimClass
    from cimprop    import CCimProp
    sresult = "component key for |%s| goes here" \
        %(propinst.sFullname)
    cAssocParent = propinst.refOwnerinst
    cAssocInstObj = cAssocParent.oInstObj
    cAssocInstInfo = cAssocInstObj.getInstInfo()
    # Not sure what order the ini file classes were in, so look
    # for the right classname here. 
    if cAssocInstInfo[0][0] == "CIM_PrintChannel":
        sConcreteClassname = cAssocInstInfo[0][0]
        nConcreteClassRowId = cAssocInstInfo[1][0]
    elif cAssocInstInfo[0][1] == "CIM_PrintChannel":
        sConcreteClassname = cAssocInstInfo[0][1]
        nConcreteClassRowId = cAssocInstInfo[1][1]
    else:
        TRC.trace(1,"error fChannelKey no Channel class in group|%s|"\
            %(cAssocInstInfo))
        return None
    TRC.trace(5,"proc  fChannelKey assoc|%s| class2|%s| rowid|%s|"\
        %(propinst.sClassname,sConcreteClassname,nConcreteClassRowId))
    sRow = nConcreteClassRowId
    spropval = fOtherClass_GetProp("InstanceID","CIM_PrintChannel",sRow)
    sresult = spropval
    return fCommonExit(propinst,myvarname,sresult,"fChannelKey")

# f I n t e r p r e t e r T y p e 
# 
def fInterpreterType(propinst,myvarname):
    from cimclass   import CCimClass
    from cimprop    import CCimProp
    cAssocParent = propinst.refOwnerinst        # class parent of property
    cAssocInstObj = cAssocParent.oInstObj       # CAssocInst instance for assoc class
    cAssocInstInfo = cAssocInstObj.getInstInfo()# list of both classnames, both rowids
    TRC.trace(1,"enter fInterpreterType inst|%s| fullname|%s|"\
        %(cAssocInstInfo,propinst.sFullname))
    # Not sure what order the ini file classes were in, so look
    # for the right classname here.  Get classname and rowid.
    if cAssocInstInfo[0][0] == "CIM_PrintChannel":
        sConcreteClassname = cAssocInstInfo[0][0]
        nChannelRowId = cAssocInstInfo[1][0]
        nInterpreterRowId = cAssocInstInfo[1][1]
    elif cAssocInstInfo[0][1] == "CIM_PrintChannel":
        sConcreteClassname = cAssocInstInfo[0][1]
        nChannelRowId = cAssocInstInfo[1][1]
        nInterpreterRowId = cAssocInstInfo[1][0]
    else:
        TRC.trace(1,"error fChannelKey no Channel class in group|%s|"\
            %(cAssocInstInfo))
        return None
    TRC.trace(5,"proc  fInterpreterType assoc|%s| class2|%s| rowid|%s|"\
        %(propinst.sClassname,sConcreteClassname,nChannelRowId))
    TRC.trace(5,"proc  fInterpreterType instinfo|%s| chanid|%s| interpid|%s|"\
        %(cAssocInstInfo,nChannelRowId,nInterpreterRowId))
    sRow = nChannelRowId
    sAssocSeq = cAssocInstObj.oGroup.sSequence

    # The parent association class retained the indices of the job control
    # language interpreter and the page description language interpreter.  
    # Arbitrarily assign one of them to each of the association instances.  
    sTypeIndex1 = cAssocParent.sChannelCurrentJobCntlLangIndex
    sTypeIndex2 = cAssocParent.sChannelDefaultPageDescLangIndex
    TRC.trace(5,"proc  fInterpreterType indices idx1|%s| idx2|%s| seq|%s|"\
        %(sTypeIndex1,sTypeIndex2,sAssocSeq))
    if str(sAssocSeq) == '1':      # check job control language
        if nInterpreterRowId == sTypeIndex1:
            spropval = 3
            TRC.trace(5,"proc  fInterpreterType match1 idx1|%s| interpid|%s| result|%s|"\
                %(sTypeIndex1,nInterpreterRowId,spropval))
        else:
            # This produces duplicate instances of the association, which
            # will have to be filtered out.  
            # If I could erase this whole instance at this late date, I would.
            # We could insert an illegal value in here that would be easy to
            # recognize by the mechanism that actually delivers the instances.
            spropval = 4
            cAssocParent.fnMarkForDeletion()
#            spropval = -1
    elif str(sAssocSeq) == '2':    # check page description language
        if nInterpreterRowId == sTypeIndex2:
            spropval = 4
            TRC.trace(5,"proc  fInterpreterType match2 idx2|%s| interpid|%s| result|%s|"\
                %(sTypeIndex2,nInterpreterRowId,spropval))
        else:
            # This produces duplicate instances of the association, which
            # will have to be filtered out.  
            # If I could erase this whole instance at this late date, I would.
            # We could insert an illegal value in here that would be easy to
            # recognize by the mechanism that actually delivers the instances.
            spropval = 3
            cAssocParent.fnMarkForDeletion()
#            spropval = -1
    else:
        spropval = 2
        TRC.trace(1,"error fInterpreterType seq|%s| jcidx|%s| pdlidx|%s| interpid|%s|"\
            %(sAssocSeq,sTypeIndex1,sTypeIndex2,nInterpreterRowId))

    # Get the InstanceID property from the right classname
    sresult = spropval
    return fCommonExit(propinst,myvarname,sresult,"fInterpreterType")


#===== A s s o c i a t e d P r i n t S u p p l y 

# f S u p p l y K e y 
# 
def fSupplyKey(propinst,myvarname):
    # First, find the correct rowid for the Supply.
    from cimclass   import CCimClass
    from cimprop    import CCimProp
    cAssocParent = propinst.refOwnerinst        # class parent of property
    cAssocInstObj = cAssocParent.oInstObj       # CAssocInst instance for assoc class
    cAssocInstInfo = cAssocInstObj.getInstInfo()# list of both classnames, both rowids
    TRC.trace(1,"enter fSupplyKey inst|%s| fullname|%s|"\
        %(cAssocInstInfo,propinst.sFullname))
    # Not sure what order the ini file classes were in, so look
    # for the right classname here.  Get classname and rowid.
    if cAssocInstInfo[0][0] == "CIM_PrintSupply":
        sConcreteClassname = cAssocInstInfo[0][0]
        nSupplyRowId = cAssocInstInfo[1][0]
        nMarkerRowId = cAssocInstInfo[1][1]
    elif cAssocInstInfo[0][1] == "CIM_PrintSupply":
        sConcreteClassname = cAssocInstInfo[0][1]
        nSupplyRowId = cAssocInstInfo[1][1]
        nMarkerRowId = cAssocInstInfo[1][0]
    else:
        TRC.trace(1,"error fSupplyKey no Supply class in group|%s|"\
            %(cAssocInstInfo))
        return None
    TRC.trace(5,"proc  fSupplyKey assoc|%s| class|%s| rowid|%s|"\
        %(propinst.sClassname,sConcreteClassname,nSupplyRowId))
    TRC.trace(5,"proc  fSupplyKey instinfo|%s| markerid|%s| supplyid|%s|"\
        %(cAssocInstInfo,nSupplyRowId,nMarkerRowId))
    sRow = nSupplyRowId
    sAssocSeq = cAssocInstObj.oGroup.sSequence
    # Next, using the rowid, get the rowid of the marker that uses it.
    sTargetMarkerIndex = cAssocParent.sSupplyMarkerIndex
    TRC.trace(5,"proc  fSupplyKey match? marker|%s| supply|%s| targetmarker|%s|"\
        %(nMarkerRowId,nSupplyRowId,sTargetMarkerIndex))
    # Finally, if we have the right marker, too, then get the InstanceID 
    # of the right Supply instance.
    if nMarkerRowId == sTargetMarkerIndex:
        itemp = nSupplyRowId
        spropval = fOtherClass_GetProp("InstanceID","CIM_PrintSupply",itemp)
        sresult = "Classname=\"%s\";InstanceID=\"%s\"" \
                    %("CIM_PrintSupply",spropval)
    else:
        sresult = "internal error: target marker index does not match marker rowid, target=|%s| marker=|%s| supply=|%s|"\
            %(sTargetMarkerIndex,nMarkerRowId,nSupplyRowId)
    return fCommonExit(propinst,myvarname,sresult,"AssociatedPrintSupply_SupplyKey")

# f M a r k e r O r F i n i s h e r K e y 
# 
# Note: Finisher is not implemented at all in this version. Therefore
# this routine always returns the key of the marker.  
def fMarkerOrFinisherKey(propinst,myvarname):
    # First, find the correct rowid for the Marker.
    from cimclass   import CCimClass
    from cimprop    import CCimProp
    cAssocParent = propinst.refOwnerinst        # class parent of property
    cAssocInstObj = cAssocParent.oInstObj       # CAssocInst instance for assoc class
    cAssocInstInfo = cAssocInstObj.getInstInfo()# list of both classnames, both rowids
    TRC.trace(1,"enter fSupplyKey inst|%s| fullname|%s|"\
        %(cAssocInstInfo,propinst.sFullname))
    # Not sure what order the ini file classes were in, so look
    # for the right classname here.  Get classname and rowid.
    if cAssocInstInfo[0][0] == "CIM_PrintSupply":
        sConcreteClassname = cAssocInstInfo[0][0]
        nSupplyRowId = cAssocInstInfo[1][0]
        nMarkerRowId = cAssocInstInfo[1][1]
    elif cAssocInstInfo[0][1] == "CIM_PrintSupply":
        sConcreteClassname = cAssocInstInfo[0][1]
        nSupplyRowId = cAssocInstInfo[1][1]
        nMarkerRowId = cAssocInstInfo[1][0]
    else:
        TRC.trace(1,"error fSupplyKey no Supply class in group|%s|"\
            %(cAssocInstInfo))
        return None
    TRC.trace(5,"proc  fSupplyKey assoc|%s| class|%s| rowid|%s|"\
        %(propinst.sClassname,sConcreteClassname,nSupplyRowId))
    TRC.trace(5,"proc  fSupplyKey instinfo|%s| markerid|%s| supplyid|%s|"\
        %(cAssocInstInfo,nSupplyRowId,nMarkerRowId))
    sRow = nSupplyRowId
    sAssocSeq = cAssocInstObj.oGroup.sSequence

    # Next, using the rowid of the supply, get the rowid of the marker that uses it.
    sTargetMarkerIndex = cAssocParent.sSupplyMarkerIndex
    TRC.trace(5,"proc  fSupplyKey match? marker|%s| supply|%s| targetmarker|%s|"\
        %(nMarkerRowId,nSupplyRowId,sTargetMarkerIndex))
    # Finally, if we have the right marker, too, then get the InstanceID 
    # of the right Marker instance.
    if nMarkerRowId == sTargetMarkerIndex:
        itemp = nMarkerRowId
        spropval = fOtherClass_GetProp("InstanceID","CIM_PrintMarker",itemp)
        sresult = "Classname=\"%s\";InstanceID=\"%s\"" \
                    %("CIM_PrintMarker",spropval)
    else:
        sresult = "internal error: target marker index does not match marker rowid, target=|%s| marker=|%s| supply=|%s|"\
            %(sTargetMarkerIndex,nMarkerRowId,nSupplyRowId)
    return fCommonExit(propinst,myvarname,sresult,"AssociatedPrintSupply_MarkerOrFinisherKey")


#end
