#!/usr/bin/python
#  provini.py                     RBLandau 20080719
# Startup phase of provider.
# 
# Copyright (c) 2008, 2009, Dell, Inc. 
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
# 
#     * Redistributions of source code must retain the above
# copyright notice, this list of conditions and the following
# disclaimer.
# 
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided
# with the distribution.
# 
#     * Neither the name of the Dell, Inc., nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
'''
This is the root of the provider.  At startup, create one instance of
the CProvider class.  It reads all the initialization files and creates
the static data structures from which the entire provider runs.  

Ini files are, in general, one line per thing, whatever the thing is.  
They are typically tab-separated, just because that makes them easier 
to parse.  Blank lines and comment lines beginning with # or ; are ignored.  
Each file is read into a single string in memory and then parsed line by
line.  All parsing is done with simple regular expressions.  

The ini files are processed in a specific order because some of them
depend on the results or previous ones.  

The processing is very stylized:
- Get the file into a list of strings (lines).  Blanks and comments are 
removed at this time.  
- Try to match each line with a regex.
- If it matches, get all the variable groupings.
- Build whatever data structure with the data.  

'''
from sys import argv, exit
from NewTrace import TRC
import re
from string import join

from dicts      import *
from printer    import *
from datastore  import datastore
from request    import *

#==============================
# C P r o v i d e r
#
# Provider object.  One instance.  
# 99% initialization.  Read all the ini files and create the static 
# data structures, mainly dictionaries, needed at runtime.  
class CProvider(object):
    '''
    Singleton class for the Provider itself, 
    primariliy initialization.
    '''
    def __init__(self):
        TRC.tracef(1,"INIT","enter CProvider init self=|"+str(self)+"|")
        pass

    # I n i t A l l ( )
    #
    # Absorb all the ini files.
    def InitAll(self,myparamdict):
        self.InitBins(myparamdict["bins"])
        self.InitOids(myparamdict["oids"])
        self.InitClassProps(myparamdict["props"])
        self.InitCli(myparamdict["cli"])
        self.InitTables(myparamdict["tables"])
        self.InitAssociations(myparamdict["assoc"])
        self.InitPrinters(myparamdict["printers"])
        self.InitAlgs(myparamdict["algs"])
        self.InitVarsubs(myparamdict["varsubs"])
        self.InitPaperSizes(myparamdict["papersizes"])
        self.InitPropEnums(myparamdict["propenums"])

    # I n i t B i n s ( )
    def InitBins(self,myfilename):
        '''Populate the cimpropname -> bin of SNMPobjectnames db'''
        # initialization: load bin lists into dict.
        TRC.tracef(1,"INIT","enter CProv InitBins file|%s|" % myfilename)
        self.lBin = self.Getfile(myfilename)
        # Extract prop name and bin list from each line.
        self.regBinProp = "^([\w\.\[\]\(\)]+)"
        self.regBinList = "[ \t]+(\w+)"
        # Get propname at the beginning of the line.
        for self.sLine in self.lBin:
            self.mMatchObj = re.search(self.regBinProp, self.sLine)
            if self.mMatchObj:
                self.sPropName = self.mMatchObj.group(0)
                # Then get the bin list of objects.
                #
                # NOTE: some bin lists may contain duplicate properties.  
                # Either filter those here before storing or be
                # very careful when retrieving them.  
                #
                self.lObjects = re.findall(self.regBinList,self.sLine)
                TRC.tracef(3,"INIT","proc  CProv InitBins prop|"+self.sPropName+"| objs|"+str(self.lObjects)+"|")
                dProp2bin.mset(self.sPropName, self.lObjects)
            else:
                TRC.tracef(1,"INIT","proc  CProv InitBins nomatch")
        TRC.tracef(2,"INIT","exit  CProv InitBins")

    # I n i t O i d s ( )
    def InitOids(self, myfilename):
        '''Populate the SNMPobjectname -> OID db and vice versa.'''
        TRC.tracef(1,"INIT","enter CProv InitOids file|%s|" % myfilename)
        self.lOid = self.Getfile(myfilename)
        # Extract prop name and OID from each line.
        self.regObjOid = "(\w+)\t+([0-9\.]+)"
        # Get propname at the beginning of the line.
        for self.sLine in self.lOid:
            TRC.tracef(5,"INIT","proc  CProv InitOids line|%s|" % self.sLine)
            self.mMatchObj = re.search(self.regObjOid, self.sLine)
            if self.mMatchObj:
                self.sObjName = self.mMatchObj.group(1)
                self.sOidNum = self.mMatchObj.group(2)
                dName2oid.mset(self.sObjName, self.sOidNum)
                dOid2name.mset(self.sOidNum, self.sObjName)
                TRC.tracef(3,"INIT","proc  CProv InitOids prop|"+self.sObjName+"| oid|"+str(self.sOidNum)+"|")
            else:
                TRC.tracef(1,"INIT","proc  CProv InitOids nomatch")
        TRC.tracef(2,"INIT","exit  CProv InitOids")

    # I n i t C l a s s P r o p s ( )
    def InitClassProps(self, myfilename):
        '''Populate the cimclassname -> list of cimpropnames db'''
        TRC.tracef(1,"INIT","enter CProv InitClassProps file|%s|" % myfilename)
        self.lPropLines = self.Getfile(myfilename)
        # Extract class name and prop list from each line.
        self.regPropClass = "^([\w\.\[\]\(\)]+)"
        self.regPropList = "\t+(\w+)"
        # Get classname at the beginning of the line.
        for self.sLine in self.lPropLines:
            TRC.tracef(5,"INIT","proc  CProv InitClassProps line|%s|" % self.sLine)
            self.mMatchObj = re.search(self.regPropClass, self.sLine)
            if self.mMatchObj:
                self.sClassName = self.mMatchObj.group(0)
                # Then get the list of properties in the class.
                self.lProps = re.findall(self.regPropList,self.sLine)
                TRC.tracef(3,"INIT","proc  CProv InitClassProps class|%s| nprops|%d| props|%s|"\
                    %(self.sClassName,len(self.lProps),str(self.lProps)) )
                dClass2props.mset(self.sClassName, self.lProps)
            else:
                TRC.tracef(1,"INIT","proc  CProv InitClassProps nomatch")
        TRC.tracef(2,"INIT","exit  CProv InitClassProps")

    # I n i t A l g s ( )
    def InitAlgs(self,myfilename):
        '''Populate the cimpropname -> algorithmfunctionname db'''
        TRC.tracef(1,"INIT","enter CProv InitAlgs file|%s|" % myfilename)
        self.lAlgs = list()
        self.lAlgs = self.Getfile(myfilename)
        # classname dot propertyname tab fCapitalizedfunctionname=varname
        self.regAlg = "^(\w+)\.(\w+)[ \t]+(f[A-Z][A-Za-z_=0-9, ]+)"
        self.regDirect = "^(\w+)=([\w, ]+)$"
        # 
        for self.sLine in self.lAlgs:
            TRC.tracef(5,"INIT","proc  CProv InitAlgs line|%s|" % self.sLine)
            self.mMatchObj = re.search(self.regAlg, self.sLine)
            if self.mMatchObj:
                self.sClassname = self.mMatchObj.group(1)
                self.sPropname = self.mMatchObj.group(2)
                self.sFuncname = self.mMatchObj.group(3)
                TRC.tracef(3,"INIT","proc  CProv InitAlgs class|%s| prop|%s| func|%s|" \
                    % (self.sClassname,self.sPropname,self.sFuncname))
                dProp2alg.mset("%s.%s"%(self.sClassname,self.sPropname),self.sFuncname)
                self.mMatchObj = None
                self.mMatchObj = re.search(self.regDirect,self.sFuncname)
                if self.mMatchObj:
                    self.sFuncname = self.mMatchObj.group(1)
                    self.sObjname = self.mMatchObj.group(2)
                    # The function takes an argument.  Register the function 
                    # and store the arg in the right dictionary for its type.
                    # This can probably be collapsed into a single dictionary
                    # in the future, because they're all indexed by classname
                    # and property name.
                    dProp2alg.mset("%s.%s"%(self.sClassname,self.sPropname),self.sFuncname)
                    fRegisterFunction(self.sFuncname)
                    fRegisterArgVar("%s.%s"%(self.sClassname,self.sPropname),self.sObjname)
                else:
                    fRegisterFunction(self.sFuncname)
        TRC.tracef(2,"INIT","exit  CProv InitAlgs")

    # I n i t P r i n t e r s ( )
    def InitPrinters(self, myfilename):
        '''Populate the list of CPrinter instances'''
        TRC.tracef(1,"INIT","enter CProv InitPrinters file|%s|" % myfilename)
        self.lPrinters = list()
        self.lPrinters = self.Getfile(myfilename)
        self.regPrinter = "^([\d\.]+)\t+(\w+)\t+(\d+)$"
        for self.sLine in self.lPrinters:
            TRC.tracef(5,"INIT","proc  CProv InitPrinters line|%s|" % self.sLine)
            self.mMatchObj = re.search(self.regPrinter, self.sLine)
            if self.mMatchObj:
                self.sIp = self.mMatchObj.group(1)
                self.sName = self.mMatchObj.group(2)
                self.sHrDevNr = self.mMatchObj.group(3)
                TRC.tracef(3,"INIT","proc  InitPrinters ip|%s| name|%s| devnr|%s|" \
                    % (self.sIp,self.sName,self.sHrDevNr))
        # TEMP TEMP TEMP 
                break
        # For the time being, there is only one printer, so we don't 
        # store a list of all the printers, or even parse more than one. 

        # Allocate a printer instance and store what we know about it
        cPrinter = CPrinter("theOnePrinter")
        cPrinter.IpAddr = self.sIp
        cPrinter.Name = self.sName
        cPrinter.SnmpHrNumber = self.sHrDevNr
        '''
        Note that this stores only the identity of the last printer.
        If there is more than one printer in the ini file, tough.  
        Can you spell P-r-o-t-o-t-y-p-e?  
        If you need more than one, make a list, and then find a way
        to pass the printer context all over the place where it is 
        needed.  
        Printer must be in the context of the individual thread.  A 
        single request might do very lengthy I/O, e.g., SNMP to a 
        device, so we cannot leave this locked over that period.  
        '''
        datastore.cPrinter = cPrinter
        # The substantive part of the printer is complete.  
        # Now build the associations.
        datastore.cPrinter.initPhase2()
        TRC.tracef(2,"INIT","exit  CProv InitPrinters")

    # I n i t T a b l e s ( )
    def InitTables(self,myfilename):
        '''
        Populate the cimclassname -> name of snmp table entry db
        and the tablename -> subscripts db
        '''
        TRC.tracef(1,"INIT","enter CProv InitTables file|%s|" % myfilename)
        self.lTableLines = self.Getfile(myfilename)
        # Extract class name and table name from each line.
        self.regTable = "^([\w_]+)\t+(\w+)\t+([\w\+]+)"
        # Get classname, at the beginning of the line, and the answer.
        for self.sLine in self.lTableLines:
            TRC.tracef(5,"INIT","proc  CProv InitTables line|%s|" % self.sLine)
            self.mMatchObj = re.search(self.regTable, self.sLine)
            if self.mMatchObj:
                self.sClassName = self.mMatchObj.group(1)
                self.sTableName = self.mMatchObj.group(2)
                self.sSubscripts = self.mMatchObj.group(3)
                TRC.tracef(3,"INIT","proc  CProv InitTables class|%s| tablename|%s| subs|%s|" \
                    % (self.sClassName,self.sTableName,self.sSubscripts))
                dClass2table.mset(self.sClassName, self.sTableName)
                dTable2subs.mset(self.sTableName, self.sSubscripts)
                # For classes that don't depend directly on a table, 
                # still write down the subscript structure for later.  
                if self.sTableName == "none":
                    dClass2subs.mset(self.sClassName, self.sSubscripts)
            else:
                TRC.tracef(1,"INIT","proc  CProv InitTables nomatch")
        TRC.tracef(2,"INIT","exit  CProv InitTables")

    # I n i t V a r s u b s ( ) 
    def InitVarsubs(self,myfilename):
        '''Populate the snmpvarname -> correction subscript specs db'''
        TRC.tracef(1,"INIT","enter CProv InitVarsubs file|%s|" % myfilename)
        self.lVarLines = self.Getfile(myfilename)
        # Extract SNMP varname subscript spec from each line.
        self.regVar = "^([\w]+)\t+([\w\+]+)"
        # Get varname at the beginning of the line.
        for self.sLine in self.lVarLines:
            TRC.tracef(5,"INIT","proc  CProv InitVarsubs line|%s|" % self.sLine)
            self.mMatchObj = re.search(self.regVar, self.sLine)
            if self.mMatchObj:
                self.sVarname = self.mMatchObj.group(1)
                self.sSubspec = self.mMatchObj.group(2)
                TRC.tracef(3,"INIT","proc  CProv InitVarsubs prop|%s| subs|%s|"%(self.sVarname,self.sSubspec))
                dVar2subs.mset(self.sVarname, self.sSubspec)
            else:
                TRC.tracef(1,"INIT","proc  CProv InitVarsubs nomatch")
        TRC.tracef(2,"INIT","exit  CProv InitVarsubs")

    # I n i t A s s o c i a t i o n s ( ) 
    def InitAssociations(self,myfilename):
        '''Populate the association name -> related classes db'''
        TRC.tracef(1,"INIT","enter CProv InitAssociations file|%s|" % myfilename)
        self.lAssociations = list()
        self.lVarLines = self.Getfile(myfilename)

        # Extract assoc name and related classes from each line.
        self.regVar = "^([\w]+)\t+([\w]+)\t+([\w]+)\t+([\w]+)\t+([\w]+)\t+([\w]+)"
        # Get varname at the beginning of the line.
        for self.sLine in self.lVarLines:
            TRC.tracef(5,"INIT","proc  CProv InitAssociations line|%s|" % self.sLine)

            self.mMatchObj = re.search(self.regVar, self.sLine)
            if self.mMatchObj:
                self.sAssocname = self.mMatchObj.group(1)
                self.sSequenceNr = self.mMatchObj.group(2)
                [self.sClassname1,self.sPropname1] = [self.mMatchObj.group(3),self.mMatchObj.group(4)]
                [self.sClassname2,self.sPropname2] = [self.mMatchObj.group(5),self.mMatchObj.group(6)]
                TRC.tracef(3,"INIT","proc  CProv InitAssociations assoc|%s| class1|%s| class2|%s|" \
                    % (self.sAssocname,self.sClassname1,self.sClassname2))
                # Assoc name is not unique, so form key from that plus related classnames.
                self.sKey = "%s.%s"%(self.sAssocname,self.sSequenceNr)
                self.lVal = [self.sClassname1,self.sClassname2]

                # Save assoc class and property structure in CAssocGroup instance.
                self.cAssocGroup = CAssocGroup(self.sAssocname,self.sSequenceNr,\
                    self.sClassname1,self.sPropname1,\
                    self.sClassname2,self.sPropname2)

                # And keep a list of the groups.
                lAssocGroups.append(self.cAssocGroup)
                # Store "assocname.classname1.classname2"=>AssocGroup
                dAssoc2groups.mset(self.sKey, self.cAssocGroup)
                # Also keep a list of associations for Request to use.
                # Keep the master list in a global place for reference.
                lAssociations.append(self.sAssocname)
                # And a parallel indexed array of the pairs of related classes.  
                lAssocClasses.append(self.lVal)
                # And this one is just a cheapo short list of association names for quick lookup.
                dAssociations.mset(self.sAssocname,self.sAssocname)
            else:
                TRC.tracef(1,"INIT","proc  CProv InitAssociations nomatch of ini line |%s|"%\
                    (self.sLine))
        TRC.tracef(2,"INIT","exit  CProv InitAssociations")

    # I n i t P a p e r S i z e s ( )
    def InitPaperSizes(self, myfilename):
        '''Populate the paper size tolerances db'''
        TRC.tracef(1,"INIT","enter CProv InitPaperSizes file|%s|" % myfilename)
        self.lPaperLines = self.Getfile(myfilename)
        # Extract size name and numbers from each line.
        self.regGetAll = "^(\w+)\t+([\d\.])+\t+(\d+)\t+(\d+)\t+(\d+)"
        for self.sLine in self.lPaperLines:
            TRC.tracef(5,"INIT","proc  CProv InitPaperSizes line|%s|" % self.sLine)
            self.mMatchObj = re.search(self.regGetAll, self.sLine)
            if self.mMatchObj:
                self.m = self.mMatchObj
                (self.sSizeName,self.sSizeNum,self.sSizeNom,self.sSizeMin,self.sSizeMax) = \
                    (self.m.group(1),self.m.group(2),self.m.group(3),self.m.group(4),self.m.group(5))
                try:
                    self.iSizeMin = int(self.sSizeMin)
                    self.iSizeMax = int(self.sSizeMax)
                except (ValueError,TypeError):
                    TRC.trace(1,"error bad int |%s| or |%s|" % \
                        (self.sSizeMin,self.sSizeMax))
                    self.iSizeMin = self.iSizeMax = 999999
                TRC.tracef(3,"INIT","proc  CProv InitPaperSizes sizename|%s| min|%d| max|%d|"\
                    %(self.sSizeName,self.iSizeMin,self.iSizeMax) )
                dPapersize2limits.mset(self.sSizeName,(self.iSizeMin,self.iSizeMax))
            else:
                TRC.tracef(1,"INIT","proc  CProv InitPaperSizes nomatch")
        TRC.tracef(2,"INIT","exit  CProv InitPaperSizes")

    # I n i t P r o p E n u m s ( ) 
    def InitPropEnums(self, myfilename):
        ''' Populate enums number -> text db'''
        TRC.tracef(1,"INIT","enter CProv InitPropEnums file|%s|" % myfilename)
        self.lEnumLines = self.Getfile(myfilename)
        # Extract enum Values and ValueMaps from each line.
        self.regGetAll = "^([\w_\.]+)\t+(\S+)\t+([\w\. -_\/,]+)"
        for self.sLine in self.lEnumLines:
            TRC.tracef(5,"INIT","proc  CProv InitPropEnums line|%s|" % self.sLine)
            self.mMatchObj = re.search(self.regGetAll, self.sLine)
            if self.mMatchObj:
                self.m = self.mMatchObj
                (self.sFullname,self.sEnumNumber,self.sEnumText) = \
                    (self.m.group(1),self.m.group(2),self.m.group(3))
                TRC.tracef(3,"INIT","proc  CProv InitPropEnums fullname|%s| val|%s| text|%s|"\
                    %(self.sFullname,self.sEnumNumber,self.sEnumText) )
                self.sFullnameKey = "%s=%s"%(self.sFullname,self.sEnumNumber)
                dPropenum2text.mset(self.sFullnameKey,self.sEnumText)
            else:
                TRC.tracef(1,"INIT","proc  CProv InitPropEnums nomatch")
        TRC.tracef(2,"INIT","exit  CProv InitPropEnums")

    # G e t f i l e ( ) 
    '''
    Helper routine to get ini file into an array of lines (strings).
    '''
    def Getfile(self,myfilename):
        TRC.tracef(1,"INIT","enter CProv Getfile file|%s|" % myfilename)
        self.lLines = list()
        if myfilename != "":
            try:
                self.fhLINES = open( myfilename, "r" )
                for self.sLine in self.fhLINES:
                    # skip blank and comment lines
                    if re.match("^\s*$",self.sLine): continue
                    if re.match("^\s*#",self.sLine): continue
                    if re.match("^\s*;",self.sLine): continue
                    # add lines to array 
                    self.lLines.append(self.sLine.strip())
            except IOError:
                print "Error: cannot open ini file " + myfilename
                exit(2)
            else:
                self.fhLINES.close()
        TRC.tracef(6,"INIT","exit  CProv Getfile="+str(len(self.lLines))+"|" + join(self.lLines,"||") + "|");
        return self.lLines


    # I n i t C l i ( )
    def InitCli(self,myfilename):
        '''
        Populate the snmpvalues db to get all printer info 
        from a file
        '''
        TRC.tracef(1,"INIT","enter CProv InitCli file|%s|" % myfilename)
        self.lCliLines = self.Getfile(myfilename)
        global lCliData
        datastore.lCliData = self.lCliLines


#==============================

